# core imports
from datetime import datetime, timedelta
import random
import math

# rcpchgrowth imports
from rcpchgrowth.global_functions import measurement_from_sds
from rcpchgrowth.measurement import Measurement

def generate_fictional_child_data(
    measurement_method: str,
    sex: str,
    start_chronological_age: float = 0.0,
    end_age: float = 20.0,
    gestation_weeks = 40,
    gestation_days = 0,
    measurement_interval_type = "days",
    measurement_interval_number: int = 20,
    start_sds = 0,
    drift = False,
    drift_range = -0.05,
    noise = False,
    noise_range = 0.01,
    reference = "uk-who"
):
  """
  This function generates an array of meassurement objects based on the params:
  measurement_method: ['height', 'weight', 'ofc', 'bmi']
  sex: ['male', 'female']
  gestation_weeks
  gestation_days,
  interval_type: ['days', 'd', 'day', 'years', 'year', 'y', 'months', 'month', 'm']
  start_sds: the starting SDS
  drift: a boolean value
  drift_range: implemented if drift is true. This is an SDS value and represents the SDS of the final plot, relative to starting SDS.
  noise: a boolean to simulate measurement accuracy
  noise_range: 0-1 - always positive. A typical acceptable error is 1% in measurement accuracy, so supplied as 0.01
  """

  # set the variables

  """
  This is an unnecessary piece of growth chart trivia included for entertainment. The first published 
  growth chart is that of the son of Count Philibert de Montbeillard (1720-1785), François Guéneau de Montbeillard.
  The date of birth used here is that of Francois.
  Acknowledgement:
  The development of growth references and growth charts, T J Cole, Ann Hum Biol. 2012 Sep; 39(5): 382–394.
  Wikipedia: https://en.wikipedia.org/wiki/Philippe_Gu%C3%A9neau_de_Montbeillard
  """
  birth_date = datetime(1759, 4, 11)  # YYYY m d
  observation_date = birth_date + timedelta(days=start_chronological_age*365.25)
  
  # set the counters
  cycle_age = start_chronological_age
  cycle_sds = start_sds 

  interval_in_years = end_age-start_chronological_age
  annualized_interval = 0 # interval between data points

  if measurement_interval_type in ['d', 'day', 'days']:
    annualized_interval = (measurement_interval_number/365.25)
  elif measurement_interval_type in ['w', 'week', 'weeks']:
    annualized_interval = (measurement_interval_number/52)
  elif measurement_interval_type in ['m', 'month', 'months']:
    annualized_interval = (measurement_interval_number/12)
  elif measurement_interval_type in ['y', 'year', 'years']:
      annualized_interval = measurement_interval_number
  else:
      raise ValueError(
          "parameters must be one of 'd', 'day', 'days', 'w', 'week', 'weeks', 'm', 'month', 'months', 'y', 'year' or 'years'")

  cycle_number = math.floor(interval_in_years/annualized_interval) # number of iterations

  drift_amount = 0.0
  if drift:
    drift_amount = drift_range / cycle_number

  measurements_array=[]

  while cycle_age < end_age:

    rawMeasurement = None

    try:
      rawMeasurement = measurement_from_sds(
        reference=reference,
        requested_sds=cycle_sds,
        measurement_method=measurement_method,
        sex=sex,
        age=cycle_age
      )
    except Exception as e:
      print(e)

    if noise and rawMeasurement is not None:
      # add measurement inaccuracy based on percentage supplied
      degree_error = rawMeasurement * noise_range
      rawMeasurement += random.uniform(-degree_error, degree_error)
    
    if rawMeasurement is not None:
      rawMeasurement = round(rawMeasurement, 1)

      measurement = Measurement(
        birth_date=birth_date,
        observation_date=observation_date,
        observation_value=rawMeasurement,
        measurement_method=measurement_method,
        reference=reference,
        sex=sex,
        gestation_weeks=gestation_weeks,
        gestation_days=gestation_days
      ).measurement

      measurements_array.append(measurement)
    
    # create drift
    if drift:
      cycle_sds += drift_amount
      # round the result
      cycle_sds=round(cycle_sds,3)
    

    # increment age
    cycle_age += annualized_interval
    observation_date = observation_date + timedelta(days=annualized_interval*365.25)
    
  return measurements_array

    


import math

"""
functions to manipulate BMI
"""

#exposed functions

"""
PARAMETERS
weight
height
bmi
ofc
"""

# public functions

def bmi_from_height_weight( height: float,  weight: float) -> float:
    """
    Returns a BMI in kg/m² from a height in cm and a weight in kg
    Does not depend on the age or sex of the child.
    Note does not test for incorrect parameters (eg height mistakenly 
    supplied in metres)
    """
    bmi = weight/math.pow(height/100, 2)
    return bmi

def weight_for_bmi_height( height: float,  bmi: float) -> float:
    """
    Returns a weight in kg from a height in cm and a BMI in kg/m²
    Does not depend on the age or sex of the child.
    Does not test for incorrect parameter values
    """
    return_weight = 0.0
    return_weight = bmi*math.pow(height/100, 2)
    return return_weight


import unittest
from datetime import date
from ..date_calculations import chronological_decimal_age, corrected_decimal_age, estimated_date_delivery


# TODO: #92 TestDecimalAge needs to be converted to use PyTest

class TestDecimalAge(unittest.TestCase):

    # Tests for term children

    def test_chronological_decimal_age_under_one_child_term(self):
        """
        calculate the decimal age of a baby under a year born at term
        """
        birth_date = date(2009, 4, 1)
        observation_date = date(2010, 2, 19)
        decimal_age = chronological_decimal_age(birth_date, observation_date)
        self.assertEqual(decimal_age, 0.8870636550308009)

    def test_chronological_decimal_age_over_two_child_term(self):
        """
        calculate the decimal age of a child over two born at term
        """
        birth_date = date(2009, 1, 3)
        observation_date = date(2011, 10, 3)
        decimal_age = chronological_decimal_age(birth_date, observation_date)
        self.assertEqual(decimal_age, 2.746064339493498)

    def test_chronological_decimal_age_over_four_child_term(self):
        """
        calculate the decimal age of a child over four born at term
        """
        birth_date = date(2009, 1, 3)
        observation_date = date(2013, 11, 20)
        decimal_age = chronological_decimal_age(birth_date, observation_date)
        self.assertEqual(decimal_age, 4.878850102669404)

    # tests from preterm children

    def test_chronological_decimal_age_premature_baby_before_term(self):
        """
        calculate the decimal age of a baby over 32 weeks not yet term
        """
        birth_date = date(2010, 12, 3)
        observation_date = date(2010, 12, 20)
        decimal_age = chronological_decimal_age(birth_date, observation_date)
        self.assertEqual(decimal_age, 0.04654346338124572)

    def test_corrected_decimal_age_premature_baby_before_term(self):
        """
        calculate the decimal age of a baby under 32 weeks not yet term
        """
        birth_date = date(2010, 12, 3)
        observation_date = date(2010, 12, 20)
        decimal_age = corrected_decimal_age(
            birth_date, observation_date, 27, 0)
        self.assertEqual(decimal_age, -0.2026009582477755)

    def test_edd_premature_baby_before_term(self):
        """
        calculate the estimated date of delivery of an baby under 32 weeks
        """
        birth_date = date(2010, 12, 3)
        edd = estimated_date_delivery(birth_date, 27, 0)
        self.assertEqual(edd, date(2011, 3, 4))


if __name__ == '__main__':
    unittest.main()



# imports from rcpchgrowth
from rcpchgrowth.constants.reference_constants import COLE_TWO_THIRDS_SDS_NINE_CENTILES, THREE_PERCENT_CENTILES
from .constants import BMI, HEAD_CIRCUMFERENCE,THREE_PERCENT_CENTILE_COLLECTION,COLE_TWO_THIRDS_SDS_NINE_CENTILE_COLLECTION 
from .global_functions import rounded_sds_for_centile

# Recommendations from Project board for reporting Centiles

# Lower limit	Upper limit	Centile band	Weight,  Height, Head	BMI
# 	<-6 		Probable error	Probable error
# -6.00	-2.84	Below 0.4th	Below normal range	Very thin
# -2.84	-2.50	0.4th 		Low BMI
# -2.50	-2.17	0.4th-2nd		Low BMI
# -2.17	-1.83	2nd		
# -1.83	-1.50	2nd -9th		
# -1.50	-1.16	9th		
# -1.16	-0.84	9th-25th		
# -0.84	-0.50	25th		
# -0.50	-0.17	25-50th 		
# -0.17	0.17	50th 		
# 0.17	0.50	50-75th 		
# 0.50	0.84	75th 		
# 0.84	1.16	75-91st 		
# 1.16	1.50	91st 		
# 1.50	1.83	91-98th 		Overweight
# 1.83	2.17	98th 		Overweight
# 2.17	2.50	98-99.6th 		Very overweight (obese)
# 2.50	2.84	99.6th 		Very overweight (obese)
# 2.84	6.00	Above 99.6th	Above normal range	Severely obese
# 	>6.00		Probable error	Probable error
"""
In fact we have changed these slightly as sometimes we might not always use the standard Cole nine
centile format. In those circumstances hard coding these thresholds would not work.
Instead @a-wei-0513 has calculated 0.25 distance between the lines to be 'on or near'. This is very close
to these numbers and would accommodate different centile formats.
"""

def return_suffix(centile: float)->str:
    # Converts a cardinal number to an ordinal by adding a suffix 'st', 'nd', 'rd' or 'th'
    # Accepts decimals and negative numbers
    
    # centile should not be < 0 or > 100
    if centile <=0:
        return "below lowest centile.;"
    if centile >= 100:
        return "above highest centile."

    suffix="th" # this is the default
    final_number = centile
    if (centile > 99 and centile <100) or (centile < 1 and centile > 0):
        final_number=round(centile,1)
    else:
       final_number=int(round(final_number))
    
    # get the final digit
    string_from_number = str(final_number)
    final_digit = int(string_from_number[len(string_from_number)-1: len(string_from_number)])
    if final_digit == 1:
        suffix = "st"
    elif final_digit == 2:
        suffix = "nd"
    elif final_digit == 3:
        suffix = "rd"
    
    # 11, 12, 13 are special cases as they take 'th'
    # get the final 2 digits if not a decimal
    if isinstance(final_number, float) and final_number.is_integer() or isinstance(final_number, int):
        final_two_digits = string_from_number[len(string_from_number)-2: len(string_from_number)]
        if int(final_two_digits) >10 <14:
            suffix = "th"

    return f"{string_from_number}{suffix}"

def quarter_distances(centile):
    """
    return sds of points that are +/- 0.25*0.666 sds space from the centile line

    0.25 * 0.666 comes from the definition that a point within 0.25 centile space
     from the centile line in a UK growth chart is considered as on the centile
    this distance from centile line is used for all centile patterns, regardless of 
     whether the centile lines are equally spaced apart
    """
    sds = rounded_sds_for_centile(centile)
    quarter_distance = 0.25 * 0.666
    return sds - quarter_distance, sds + quarter_distance


def generate_centile_band_ranges(centile_collection):
    """
    returns a list of tuples representing ranges of on centile line and between centiles
    """
    centile_ranges = []
    for centile in centile_collection:
        centile_ranges.extend(quarter_distances(centile))
    centile_bands = list(zip(centile_ranges[:-1],centile_ranges[1:]))

    return centile_bands


def centile_band_for_centile(sds: float, measurement_method: str, centile_format: str)->str:
    """
        this function returns a centile band into which the sds falls
    
        params: accepts a sds: float
        params: accepts a measurement_method as string
        params: accepts array of centiles representing the centile lines
    """
    
    centile_collection = []
    if centile_format == THREE_PERCENT_CENTILES:
        centile_collection = THREE_PERCENT_CENTILE_COLLECTION
    elif centile_format == COLE_TWO_THIRDS_SDS_NINE_CENTILES:
        centile_collection = COLE_TWO_THIRDS_SDS_NINE_CENTILE_COLLECTION

    centile_band_ranges = generate_centile_band_ranges(centile_collection)

    if measurement_method == BMI:
        measurement_method = "body mass index"
    elif measurement_method == HEAD_CIRCUMFERENCE:
        measurement_method = "head circumference"

    if sds <= -6:
        return f"This {measurement_method} measurement is well below the normal range. Please review its accuracy."
    elif sds > 6:
        return f"This {measurement_method} measurement is well above the normal range. Please review its accuracy."
    elif sds <= centile_band_ranges[0][0]:
        return f"This {measurement_method} measurement is below the normal range"
    elif sds > centile_band_ranges[-1][1]:
        return f"This {measurement_method} measurement is above the normal range"        
    else:
        #even indices of centile_bands list is always on centile
        #odd indices of cnetile_bands list is always between centiles
        for r in range(len(centile_band_ranges)):
            if centile_band_ranges[r][0] <= sds < centile_band_ranges[r][1]:
                if r%2 == 0:
                    centile = centile_collection[r//2]
                    suffixed_centile = return_suffix(centile)
                    return f"This {measurement_method} measurement is on or near the {suffixed_centile} centile."
                else:
                    lower_centile = centile_collection[(r-1)//2]
                    lower_suffixed_centile = return_suffix(lower_centile)
                    upper_centile = centile_collection[(r+1)//2]
                    upper_suffixed_centile = return_suffix(upper_centile)
                    return f"This {measurement_method} measurement is between the {lower_suffixed_centile} and {upper_suffixed_centile} centiles."



from typing import Union
from .global_functions import centile, lms_value_array_for_measurement_for_reference, sds_for_centile, rounded_sds_for_centile, generate_centile
from .uk_who import select_reference_data_for_uk_who_chart
from .trisomy_21 import select_reference_data_for_trisomy_21
from .turner import select_reference_data_for_turner
from .constants.reference_constants import FEMALE, HEIGHT, UK_WHO, TURNERS, TRISOMY_21, COLE_TWO_THIRDS_SDS_NINE_CENTILES, COLE_TWO_THIRDS_SDS_NINE_CENTILE_COLLECTION, THREE_PERCENT_CENTILE_COLLECTION, UK_WHO_REFERENCES

"""
Public chart functions
"""


def create_chart(reference: str, centile_format: Union[str, list] = COLE_TWO_THIRDS_SDS_NINE_CENTILES, measurement_method: str = HEIGHT, sex: str = FEMALE, is_sds=False):
    """
    Global method - return chart for measurement_method, sex and reference
    """
    
    if reference == UK_WHO:
        return create_uk_who_chart(measurement_method=measurement_method, sex=sex, centile_format=centile_format, is_sds=is_sds)
    elif reference == TURNERS:
        return create_turner_chart(centile_format=centile_format, is_sds=is_sds)
    elif reference == TRISOMY_21:
        return create_trisomy_21_chart(measurement_method=measurement_method, sex=sex, centile_format=centile_format, is_sds=is_sds)
    else:
        print("No reference data returned. Is there a spelling mistake in your reference?")

def generate_custom_sds_line(reference: str, measurement_method: str, sex: str, custom_sds: float):
    # Public function that returns a custom SDS line
    # the centile reference data
    
    ##
    # iterate through the 4 references that make up UK-WHO
    # There will be a list for each one. For the other references ther will be only one list
    ##

    # all data for a given reference are stored here: this is returned to the user
    reference_data = []

    custom_centile = centile(custom_sds)

    if reference == UK_WHO:
        for reference_index, reference in enumerate(UK_WHO_REFERENCES):
            # the centile reference data
            lms_array_for_measurement=select_reference_data_for_uk_who_chart(
                uk_who_reference=reference, 
                measurement_method=measurement_method, 
                sex=sex)
            centile_data=[]
            try:
                centile_data= build_centile_object(
                    measurement_method=measurement_method,
                    sex=sex,
                    lms_array_for_measurement=lms_array_for_measurement,
                    z=custom_sds,
                    centile=custom_centile
                )
            except:
                print("Could not generate centile data.")
                centile_data=[]
        # all data can now be tagged by reference_name and added to reference_data
        reference_data.append({reference: centile_data})
    else:
        # get the reference data
        lms_array_for_measurement=[]
        try:
            lms_array_for_measurement=select_reference_lms_data(
                reference=reference,
                measurement_method=measurement_method,
                sex=sex
            )
        except:
            print(f"It was not possible to retrieve {reference} data.")
            lms_array_for_measurement=[]

        try:
            centile_data=[]
            centile_data= build_centile_object(
                    measurement_method=measurement_method,
                    sex=sex,
                    lms_array_for_measurement=lms_array_for_measurement,
                    z=custom_sds,
                    centile=custom_centile
                )
        except:
            print("Could not generate sds data.")
            centile_data=[]

        reference_data.append({reference: centile_data})

    return reference_data

def generate_custom_centile(reference: str, measurement_method: str, sex: str, custom_centile: float)->list:

    # Public function that returns a custom centile line
    # the centile reference data
    
    ##
    # iterate through the 4 references that make up UK-WHO
    # There will be a list for each one. For the other references ther will be only one list
    ##

    # all data for a given reference are stored here: this is returned to the user
    reference_data = []

    z = sds_for_centile(custom_centile)

    if reference == UK_WHO:
        for reference_index, reference in enumerate(UK_WHO_REFERENCES):
            # the centile reference data
            lms_array_for_measurement=select_reference_data_for_uk_who_chart(
                uk_who_reference=reference, 
                measurement_method=measurement_method, 
                sex=sex)
            centile_data=[]
            try:
                centile_data= build_centile_object(
                    measurement_method=measurement_method,
                    sex=sex,
                    lms_array_for_measurement=lms_array_for_measurement,
                    z=z,
                    centile=custom_centile
                )
            except:
                print("Could not generate centile data.")
                centile_data=[]
        # all data can now be tagged by reference_name and added to reference_data
        reference_data.append({reference: centile_data})
    else:
        # get the reference data
        lms_array_for_measurement=[]
        try:
            lms_array_for_measurement=select_reference_lms_data(
                reference=reference,
                measurement_method=measurement_method,
                sex=sex
            )
        except:
            print(f"It was not possible to retrieve {reference} data.")
            lms_array_for_measurement=[]

        try:
            centile_data=[]
            centile_data= build_centile_object(
                    measurement_method=measurement_method,
                    sex=sex,
                    lms_array_for_measurement=lms_array_for_measurement,
                    z=z,
                    centile=custom_centile
                )
        except:
            print("Could not generate centile data.")
            centile_data=[]

        reference_data.append({reference: centile_data})

    return reference_data


def create_plottable_child_data(child_results_array):
    """
    DEPRECATED
    Global method - receives a measurement object and returns the data in plottable format - the ages as x, the measurements
    as y and the centile values as l (label)
    """

    centile_data = []
    sds_data = []

    for count, child_result in enumerate(child_results_array):
        if(child_result):
            # create 4 plottable return objects for each measurement: one for each corrected and chronological age
            # per measurement and per SDS score
            # These measurement pairs and SDS pairs are stored in a 2 2-value arrays, so that each measurement/SDS
            # can be plotted as a series in the charts.
            # If there are multiple values to plot, the return array will be a string of arrays of paired values,
            # which allows them to be plotted as pairs: this is because corrected and chronological values should be
            # linked by a line, the chronological value denotes as a dot, the corrected value as a cross.

            chronological_data_point = {
                "measurement_method": child_result["child_observation_value"]["measurement_method"],
                "x": child_result["measurement_dates"]["chronological_decimal_age"],
                "y": child_result["child_observation_value"]["observation_value"],
                "observation_value_error": child_result["child_observation_value"]["observation_value_error"],
                "centile_band": child_result["measurement_calculated_values"]["chronological_centile_band"],
                "centile_value": child_result["measurement_calculated_values"]["chronological_centile"],
                "sds": child_result["measurement_calculated_values"]["chronological_sds"],
                "measurement_error": child_result["measurement_calculated_values"]["chronological_measurement_error"],
                "age_error": child_result["measurement_dates"]["chronological_decimal_age_error"],
                "age_type": "chronological_age",
                "calendar_age": child_result["measurement_dates"]["chronological_calendar_age"],
                "corrected_gestation_weeks": child_result["measurement_dates"]["corrected_gestational_age"]["corrected_gestation_weeks"],
                "corrected_gestation_days": child_result["measurement_dates"]["corrected_gestational_age"]["corrected_gestation_days"],
                "lay_chronological_decimal_age_comment": child_result["measurement_dates"]["comments"]["lay_chronological_decimal_age_comment"],
                "clinician_chronological_decimal_age_comment": child_result["measurement_dates"]["comments"]["clinician_chronological_decimal_age_comment"]
            }
            corrected_data_point = {
                "measurement_method": child_result["child_observation_value"]["measurement_method"],
                "x": child_result["measurement_dates"]["corrected_decimal_age"],
                "y": child_result["child_observation_value"]["observation_value"],
                "observation_value_error": child_result["child_observation_value"]["observation_value_error"],
                "centile_band": child_result["measurement_calculated_values"]["corrected_centile_band"],
                "centile_value": child_result["measurement_calculated_values"]["corrected_centile"],
                "sds": child_result["measurement_calculated_values"]["corrected_sds"],
                "measurement_error": child_result["measurement_calculated_values"]["corrected_measurement_error"],
                "age_error": child_result["measurement_dates"]["corrected_decimal_age_error"],
                "age_type": "corrected_age",
                "calendar_age": child_result["measurement_dates"]["corrected_calendar_age"],
                "corrected_gestation_weeks": child_result["measurement_dates"]["corrected_gestational_age"]["corrected_gestation_weeks"],
                "corrected_gestation_days": child_result["measurement_dates"]["corrected_gestational_age"]["corrected_gestation_days"],
                "lay_corrected_decimal_age_comment": child_result["measurement_dates"]["comments"]["lay_corrected_decimal_age_comment"],
                "clinician_corrected_decimal_age_comment": child_result["measurement_dates"]["comments"]["clinician_corrected_decimal_age_comment"],
            }
            chronological_sds_data_point = {
                "measurement_method": child_result["child_observation_value"]["measurement_method"],
                "x": child_result["measurement_dates"]["chronological_decimal_age"],
                "y": child_result["measurement_calculated_values"]["chronological_sds"],
                "observation_value_error": child_result["child_observation_value"]["observation_value_error"],
                "sds": child_result["measurement_calculated_values"]["chronological_sds"],
                "measurement_error": child_result["measurement_calculated_values"]["chronological_measurement_error"],
                "age_error": child_result["measurement_dates"]["chronological_decimal_age_error"],
                "age_type": "chronological_age",
                "calendar_age": child_result["measurement_dates"]["chronological_calendar_age"],
                "corrected_gestation_weeks": child_result["measurement_dates"]["corrected_gestational_age"]["corrected_gestation_weeks"],
                "corrected_gestation_days": child_result["measurement_dates"]["corrected_gestational_age"]["corrected_gestation_days"],
                "lay_chronological_decimal_age_comment": child_result["measurement_dates"]["comments"]["lay_chronological_decimal_age_comment"],
                "clinician_chronological_decimal_age_comment": child_result["measurement_dates"]["comments"]["clinician_chronological_decimal_age_comment"]
            }

            corrected_sds_data_point = {
                "measurement_method": child_result["child_observation_value"]["measurement_method"],
                "x": child_result["measurement_dates"]["corrected_decimal_age"],
                "y": child_result["measurement_calculated_values"]["corrected_sds"],
                "observation_value_error": child_result["child_observation_value"]["observation_value_error"],
                "sds": child_result["measurement_calculated_values"]["corrected_sds"],
                "measurement_error": child_result["measurement_calculated_values"]["corrected_measurement_error"],
                "age_error": child_result["measurement_dates"]["corrected_decimal_age_error"],
                "age_type": "corrected_age",
                "calendar_age": child_result["measurement_dates"]["corrected_calendar_age"],
                "corrected_gestation_weeks": child_result["measurement_dates"]["corrected_gestational_age"]["corrected_gestation_weeks"],
                "corrected_gestation_days": child_result["measurement_dates"]["corrected_gestational_age"]["corrected_gestation_days"],
                "lay_corrected_decimal_age_comment": child_result["measurement_dates"]["comments"]["lay_corrected_decimal_age_comment"],
                "clinician_corrected_decimal_age_comment": child_result["measurement_dates"]["comments"]["clinician_corrected_decimal_age_comment"],
            }

            measurement_data_points = [
                corrected_data_point, chronological_data_point]
            measurement_sds_data_points = [
                corrected_sds_data_point, chronological_sds_data_point]

            centile_data.append(measurement_data_points)
            sds_data.append(measurement_sds_data_points)

    result = {
        "measurement_method": child_result["child_observation_value"]["measurement_method"],
        "centile_data": centile_data,
        "sds_data": sds_data
    }

    return result

    """
    Return object structure

    [
        heights: [
            [{
                x: 9.415, `this is the corrected age of the child at date of measurement in decimal years
                y: 120 `this is the observation value - the units will be added in the client
                "centile_band": 'Your child's height is between the 75th and 91st centiles' `a text advice string for labelling - corrected,
                "centile_value": 86 `centile number value - reported but not used: the project board do not like exact centile numbers - corrected,
                "age_type": "corrected_age", `this is a flag to differentiate the two ages for the same observation_value
                "calendar_age": "9 years and 4 months" `this is the calendar age for labelling
                "corrected_gestational_weeks": 23 `the corrected gestational age if relevant for labelling
                "corrected_gestational_days": 1 `the corrected gestational age if relevant for labelling
            }
            {
                x: 9.415, `this is the chronological age of the child at date of measurement in decimal years
                y: 120 `this is the observation value - the units will be added in the client
                "centile_band": 'Your child's height is between the 75th and 91st centiles' `a text advice string for labelling - based on chronological age,
                "centile_value": 86 `centile number value - reported but not used: the project board do not like exact centile numbers - based on chronological age, 
                "age_type": "corrected_age", `this is a flag to differentiate the two ages for the same observation_value
                "calendar_age": "9 years and 4 months" `this is the calendar age for labelling
                "corrected_gestational_weeks": 23 `the corrected gestational age if relevant for labelling
                "corrected_gestational_days": 1 `the corrected gestational age if relevant for labelling
            }
            ]
        ],
        height_sds: [
                x: 9.415, `this is the age of the child at date of measurement in decimal years
                y: 1.3 `this is the SDS value for SDS charts
                "age_type": "corrected_age", `this is a flag to differentiate the two ages for the same observation_value
                "calendar_age": "9 years and 4 months" `this is the calendar age for labelling
                "corrected_gestational_weeks": 23 `the corrected gestational age if relevant for labelling
                "corrected_gestational_days": 1 `the corrected gestational age if relevant for labelling
        ],
        .... and so on for the other measurement_methods
        
    ]

    """


"""
private functions
"""
def select_reference_lms_data(reference: str, measurement_method: str, sex: str)->list:
    lms_array_for_measurement = []
    if reference == TURNERS:
        lms_array_for_measurement=select_reference_data_for_turner(measurement_method=measurement_method, sex=sex)
    elif reference == TRISOMY_21:
        lms_array_for_measurement=select_reference_data_for_trisomy_21(measurement_method=measurement_method, sex=sex)
    else: 
        raise Exception("No data has been selected!")
    
    return lms_array_for_measurement
    


def build_centile_object(measurement_method: str, sex: str, lms_array_for_measurement: list, z: float, centile: float):
    sex_list: dict = {}  # all the data for a given sex are stored here
    measurements: dict = {}  # all the data for a given measurement_method are stored here
    centiles = []  # all generated centiles for a selected centile collection are stored here

    try:
        # Generate a centile. there will be nine of these if Cole method selected.
        # Some data does not exist at all ages, so any error reflects missing data.
        # If this happens, an empty list is returned.
        centile_data = generate_centile(
            z=z,
            centile=centile,
            measurement_method=measurement_method,
            sex=sex,
            lms_array_for_measurement=lms_array_for_measurement,
            reference=UK_WHO
        )
    except:
        centile_data=None
    # Store this centile for a given measurement
    
    centiles.append({"sds": round(z * 100) / 100,
                "centile": centile, "data": centile_data})

    # this is the end of the centile_collection for loop
    # All the centiles for this measurement, sex and reference are added to the measurements list
    measurements.update({measurement_method: centiles})

    # this is the end of the measurement_methods loop
    # All data for all measurement_methods for this sex are added to the sex_list list

    sex_list.update({sex: measurements})

    return sex_list


def create_uk_who_chart(measurement_method: str, sex: str, centile_format: Union[str, list] = COLE_TWO_THIRDS_SDS_NINE_CENTILES, is_sds = False):

    # user selects which centile collection they want, for sex and measurement_method
    # If the Cole method is selected, conversion between centile and SDS
    # is different as SDS is rounded to the nearest 2/3
    # Cole method selection is stored in the cole_method flag.
    # If no parameter is passed, default is the Cole method
    # Alternatively it is possible to pass a custom list of values - if the is_sds flag is False (default) these are centiles

    centile_sds_collection = []
    cole_method = False

    if (type(centile_format) is list):
        centile_sds_collection = centile_format
    elif centile_format == COLE_TWO_THIRDS_SDS_NINE_CENTILES:
        centile_sds_collection = COLE_TWO_THIRDS_SDS_NINE_CENTILE_COLLECTION
        cole_method = True
        is_sds=False
    else:
        centile_sds_collection = THREE_PERCENT_CENTILE_COLLECTION
        is_sds=False
    ##
    # iterate through the 4 references that make up UK-WHO
    # There will be a list for each one
    ##

    # all data for a given reference are stored here: this is returned to the user
    reference_data = []

    for reference_index, reference in enumerate(UK_WHO_REFERENCES):
        sex_list: dict = {}  # all the data for a given sex are stored here
        # For each reference we have 2 sexes
        # for sex_index, sex in enumerate(SEXES):
        # For each sex we have 4 measurement_methods

        measurements: dict = {}  # all the data for a given measurement_method are stored here

        # for measurement_index, measurement_method in enumerate(MEASUREMENT_METHODS):
        # for every measurement method we have as many centiles
        # as have been requested

        centiles = []  # all generated centiles for a selected centile collection are stored here

        # the centile reference data
        try:
            lms_array_for_measurement=select_reference_data_for_uk_who_chart(
        uk_who_reference=reference, measurement_method=measurement_method, sex=sex)
        except:
            lms_array_for_measurement = []

        for centile_index, centile_sds in enumerate(centile_sds_collection):
            # we must create a z for each requested centile
            # if the Cole 9 centiles were selected, these are rounded,
            # so conversion to SDS is different
            # Otherwise standard conversation of centile to z is used

            z=0.0 #initialise
            centile_value=0.0 #initialise

            if cole_method:
                z = rounded_sds_for_centile(centile_sds) # a centile was provided, so convert to z
                centile_value=centile_sds # store the original centile value 
            else:
                if (is_sds):
                    z=centile_sds # an sds was supplied
                    centile_value=centile(centile_sds) # convert the z to a centile and store
                else:
                    z = sds_for_centile(centile_sds) # a centile was provided, so convert to z
                    centile_value=centile_sds # store the original centile value 
            centile_data = []

            try:
                # Generate a centile. there will be nine of these if Cole method selected.
                # Some data does not exist at all ages, so any error reflects missing data.
                # If this happens, an empty list is returned.
                centile_data = generate_centile(
                    z=z,
                    centile=centile_value,
                    measurement_method=measurement_method,
                    sex=sex,
                    lms_array_for_measurement=lms_array_for_measurement,
                    reference=UK_WHO,
                    is_sds=is_sds
                )
            except:
                print(f"Not possible to generate centile data for UK-WHO {measurement_method} in {sex}s.")
                centile_data=None
            # Store this centile for a given measurement
            
            centiles.append({"sds": round(z * 100) / 100,
                        "centile": centile_value, "data": centile_data})

        # this is the end of the centile_collection for loop
        # All the centiles for this measurement, sex and reference are added to the measurements list
        measurements.update({measurement_method: centiles})

        # this is the end of the measurement_methods loop
        # All data for all measurement_methods for this sex are added to the sex_list list

        sex_list.update({sex: measurements})

        # all data can now be tagged by reference_name and added to reference_data
        reference_data.append({reference: sex_list})

    # returns a list of 4 references, each containing 2 lists for each sex,
    # each sex in turn containing 4 datasets for each measurement_method
    return reference_data

    """
    structure:
    UK_WHO generates 4 json objects, each structure as below
    uk90_preterm: {
        male: {
            height: [
                {
                    sds: -2.667,
                    centile: 0.4
                    data: [{l: , x: , y: }, ....]
                }
            ],
            weight: [...]
        },
        female {...}
    }
    uk_who_infant: {...}
    uk_who_child:{...}
    uk90_child: {...}
    """


def create_turner_chart(centile_format: Union[str, list], is_sds=False):
   # user selects which centile collection they want
    # If the Cole method is selected, conversion between centile and SDS
    # is different as SDS is rounded to the nearest 2/3
    # Cole method selection is stored in the cole_method flag.
    # If no parameter is passed, default is the Cole method

    centile_sds_collection = []
    cole_method = False

    if (type(centile_format) is list):
        centile_sds_collection = centile_format
    elif centile_format == COLE_TWO_THIRDS_SDS_NINE_CENTILES:
        centile_sds_collection = COLE_TWO_THIRDS_SDS_NINE_CENTILE_COLLECTION
        cole_method = True
        is_sds=False
    else:
        centile_sds_collection = THREE_PERCENT_CENTILE_COLLECTION
        is_sds=False

    # all data for a the reference are stored here: this is returned to the user
    reference_data = {}

    # for sex_index, sex in enumerate(SEXES):
    # For each sex we have 4 measurement_methods
    # Turner is female only, but we will generate empty arrays for male
    # data to keep all objects the same

    sex_list: dict = {}
    sex = FEMALE

    measurements: dict = {}  # all the data for a given measurement_method are stored here

    # for measurement_index, measurement_method in enumerate(MEASUREMENT_METHODS):
    # for every measurement method we have as many centiles
    # as have been requested

    centiles = []  # all generated centiles for a selected centile collection are stored here

    for centile_index, centile_sds in enumerate(centile_sds_collection):
        # we must create a z for each requested centile
        # if the Cole 9 centiles were selected, these are rounded,
        # so conversion to SDS is different
        # Otherwise standard conversation of centile to z is used
        if cole_method:
            z = rounded_sds_for_centile(centile_sds)
            centile_value=centile_sds
        else:
            if is_sds:
                z = centile_sds
                centile_value = centile(centile_sds)
            else:
                z = sds_for_centile(centile_sds)
                centile_value=centile_sds
        # Collect the LMS values from the correct reference
        lms_array_for_measurement = select_reference_data_for_turner(
            measurement_method=HEIGHT, sex=sex)
        # Generate a centile. there will be nine of these if Cole method selected.
        # Some data does not exist at all ages, so any error reflects missing data.
        # If this happens, an empty list is returned.
        try:
            centile_data = generate_centile(
                z=z, 
                centile=centile_value, 
                measurement_method=HEIGHT,
                sex=sex, 
                lms_array_for_measurement=lms_array_for_measurement, reference=TURNERS,
                is_sds=is_sds)

            # Store this centile for a given measurement
            centiles.append({"sds": round(z * 100) / 100,
                            "centile": centile_value, "data": centile_data})
        except Exception as e:
            print(e)

    # this is the end of the centile_collection for loop
    # All the centiles for this measurement, sex and reference are added to the measurements list
    measurements.update({HEIGHT: centiles})

    # this is the end of the measurement_methods loop
    # All data for all measurement_methods for this sex are added to the sex_list list

    sex_list.update({sex: measurements})

    # all data can now be tagged by reference_name and added to reference_data
    reference_data = [{TURNERS: sex_list}]
    return reference_data

    """
    Return object structure
    [turners-syndrome: [{
        female: {
            height: [
                {
                    sds: -2.667,
                    centile: 0.4
                    data: [{l: , x: , y: }, ....]
                }
            ],
            weight: [...]
        },
        female {...}
    }]
    """


def create_trisomy_21_chart(measurement_method: str, sex: str, centile_format: Union[str, list], is_sds=False):
   # user selects which centile collection they want
    # If the Cole method is selected, conversion between centile and SDS
    # is different as SDS is rounded to the nearest 2/3
    # Cole method selection is stored in the cole_method flag.
    # If no parameter is passed, default is the Cole method

    centile_sds_collection = []
    cole_method = False

    if (type(centile_format) is list):
        centile_sds_collection = centile_format
    elif centile_format == COLE_TWO_THIRDS_SDS_NINE_CENTILES:
        centile_sds_collection = COLE_TWO_THIRDS_SDS_NINE_CENTILE_COLLECTION
        cole_method = True
        is_sds=False
    else:
        centile_sds_collection = THREE_PERCENT_CENTILE_COLLECTION
        is_sds=False

    # all data for a the reference are stored here: this is returned to the user
    reference_data = {}
    sex_list: dict = {}

    # for sex_index, sex in enumerate(SEXES):
    # For each sex we have 4 measurement_methods

    measurements: dict = {}  # all the data for a given measurement_method are stored here

    # for measurement_index, measurement_method in enumerate(MEASUREMENT_METHODS):
    # for every measurement method we have as many centiles
    # as have been requested

    centiles = []  # all generated centiles for a selected centile collection are stored here

    for centile_index, centile_sds in enumerate(centile_sds_collection):
        # we must create a z for each requested centile
        # if the Cole 9 centiles were selected, these are rounded,
        # so conversion to SDS is different
        # Otherwise standard conversation of centile to z is used
        if cole_method:
            z = rounded_sds_for_centile(centile_sds)
            centile_value=centile_sds
        else:
            if is_sds:
                z = centile_sds
                centile_value=centile(z)
            else:
                z = sds_for_centile(centile_sds)
                centile_value=centile_sds
        # Collect the LMS values from the correct reference
        lms_array_for_measurement = select_reference_data_for_trisomy_21(
            measurement_method=measurement_method, sex=sex)
        # Generate a centile. there will be nine of these if Cole method selected.
        # Some data does not exist at all ages, so any error reflects missing data.
        # If this happens, an empty list is returned.
        try:    
            centile_data = generate_centile(
                z=z, 
                centile=centile_value, 
                measurement_method=measurement_method,
                sex=sex, 
                lms_array_for_measurement=lms_array_for_measurement, 
                reference=TRISOMY_21,
                is_sds=is_sds)

            # Store this centile for a given measurement
            centiles.append({"sds": round(z, 2),
                            "centile": centile_value, "data": centile_data})
        except Exception as e:
            print(f"{e}")

    # this is the end of the centile_collection for loop
    # All the centiles for this measurement, sex and reference are added to the measurements list
    measurements.update({measurement_method: centiles})

    # this is the end of the measurement_methods loop
    # All data for all measurement_methods for this sex are added to the sex_list list

    sex_list.update({sex: measurements})

    # all data can now be tagged by reference_name and added to reference_data
    reference_data = [{TRISOMY_21: sex_list}]
    return reference_data

    """
    # return object structure
    [trisomy_21: {
        male: {
            height: [
                {
                    sds: -2.667,
                    centile: 0.4
                    data: [{l: , x: , y: }, ....]
                }
            ],
            weight: [...]
        },
        female {...}
    }]
    """



"""
Handles UK-WHO-specific reference data selection
"""

# standard imports
import json
import pkg_resources

# rcpch imports
from .constants import *

"""
birth_date: date of birth
observation_date: date of observation
sex: sex (string, MALE or FEMALE)
decimal_age: chronological, decimal
corrected_age: corrected for prematurity, decimal
measurement_method: height, weight, bmi, ofc (decimal)
observation: value (float)
gestation_weeks: gestational age(weeks), integer
gestation_days: supplementary days of gestation
lms: L, M or S
reference: reference data
"""

# load the reference data

UK90_PRETERM_DATA = pkg_resources.resource_filename(
    __name__, "/data_tables/uk90_preterm.json")  # 23 - 42 weeks gestation
with open(UK90_PRETERM_DATA) as json_file:
    UK90_PRETERM_DATA = json.load(json_file)
    json_file.close()

UK90_TERM_DATA = pkg_resources.resource_filename(
    __name__, "/data_tables/uk90_term.json")  # 37-42 weeks gestation
with open(UK90_TERM_DATA) as json_file:
    UK90_TERM_DATA = json.load(json_file)
    json_file.close()

WHO_INFANTS_DATA = pkg_resources.resource_filename(
    __name__, "/data_tables/who_infants.json")  # 2 weeks to 2 years
with open(WHO_INFANTS_DATA) as json_file:
    WHO_INFANTS_DATA = json.load(json_file)
    json_file.close()

WHO_CHILD_DATA = pkg_resources.resource_filename(
    __name__, "/data_tables/who_children.json")  # 2 years to 4 years
with open(WHO_CHILD_DATA) as json_file:
    WHO_CHILD_DATA = json.load(json_file)
    json_file.close()

UK90_CHILD_DATA = pkg_resources.resource_filename(
    __name__, "/data_tables/uk90_child.json")  # 4 years to 20 years
with open(UK90_CHILD_DATA) as json_file:
    UK90_CHILD_DATA = json.load(json_file)
    json_file.close()

# public functions


def reference_data_absent(
    age: float,
    measurement_method: str,
    sex: str
):
    """
    Helper function.
    Returns boolean
    Tests presence of valid reference data for a given measurement request

    Reference data is not complete for all ages/sexes/measurements.
     - Length data is not available until 25 weeks gestation, though weight date is available from 23 weeks
     - There is only BMI reference data from 2 weeks of age to aged 20y
     - Head circumference reference data is available from 23 weeks gestation to 17y in girls and 18y in boys
     - lowest threshold is 23 weeks, upper threshold is 20y
    """

    if age < TWENTY_THREE_WEEKS_GESTATION:  # lower threshold of UK90 data
        return True, "UK-WHO data does not exist below 23 weeks gestation."

    if age > TWENTY_YEARS:  # upper threshold of UK90 data
        return True, "UK-WHO data does not exist above 20 years."

    if measurement_method == HEIGHT and age < TWENTY_FIVE_WEEKS_GESTATION:
        return True, "UK-WHO length data does not exist in infants below 25 weeks gestation."

    elif measurement_method == BMI and age < FORTY_TWO_WEEKS_GESTATION:
        return True, "UK-WHO BMI data does not exist below 2 weeks of age."

    elif measurement_method == HEAD_CIRCUMFERENCE:
        if (sex == MALE and age > EIGHTEEN_YEARS) or (sex == FEMALE and age > SEVENTEEN_YEARS):
            if sex == MALE:
                return True, "UK-WHO head circumference data does not exist in boys over 18 y of age."
            else:
                return True, "UK-WHO head circumference data does not exist in girls over 17 y of age."
        else:
            return False, ""
    else:
        return False, ""


def uk_who_reference(
    age: float
) -> json:
    """
    The purpose of this function is to choose the correct reference for calculation.
    The UK-WHO standard is an unusual case because it combines two different reference sources.
    - UK90 reference runs from 23 weeks to 20 y
    - WHO 2006 runs from 2 weeks to 4 years
    - UK90 then resumes from 4 years to 20 years
    The function return the appropriate reference file as json
    """

    # These conditionals are to select the correct reference
    if age < UK90_REFERENCE_LOWER_THRESHOLD:
        # Below the range for which we have reference data, we can't provide a calculation.
        return ValueError("There is no UK90 reference data below 23 weeks gestation")
    elif age < UK_WHO_INFANT_LOWER_THRESHOLD:
        # Below 42 weeks, the UK90 preterm data is always used
        return UK90_PRETERM_DATA

    elif age < WHO_CHILD_LOWER_THRESHOLD:
        # Children beyond 2 weeks but below 2 years are measured lying down using WHO data
        return WHO_INFANTS_DATA

    elif age < WHO_CHILDREN_UPPER_THRESHOLD:
        # Children 2 years and beyond but below 4 years are measured standing up using WHO data        
        return WHO_CHILD_DATA
        
    elif age <= UK90_UPPER_THRESHOLD:
        # All children 4 years and above are measured using UK90 child data
        return UK90_CHILD_DATA

    else:
        return ValueError("There is no UK90 reference data above the age of 20 years.")


def uk_who_lms_array_for_measurement_and_sex(
    age: float,
    measurement_method: str,
    sex: str
) -> list:

    # selects the correct lms data array from the patchwork of references that make up UK-WHO

    try:
        selected_reference = uk_who_reference(
            age=age
        )
    except:  #  there is no reference for the age supplied
        return LookupError("There is no UK-WHO reference for the age supplied.")

    # Check that the measurement requested has reference data at that age

    invalid_data, data_error = reference_data_absent(
        age=age,
        measurement_method=measurement_method,
        sex=sex)

    if invalid_data:
        raise LookupError(data_error)
    else:
        return selected_reference["measurement"][measurement_method][sex]


def select_reference_data_for_uk_who_chart(
    uk_who_reference: str, 
    measurement_method: str, 
    sex: str):

    # takes a uk_who_reference name (see parameter constants), measurement_method and sex to return
    # reference data

    if uk_who_reference == UK90_PRETERM:
        try:
            uk90_preterm_reference = uk_who_lms_array_for_measurement_and_sex(
                age=-0.01,
                measurement_method=measurement_method,
                sex=sex,
            )
        except:
            uk90_preterm_reference = []
        return uk90_preterm_reference
    elif uk_who_reference == UK_WHO_INFANT:
        try:
            uk_who_infants_reference = uk_who_lms_array_for_measurement_and_sex(
                age=0.04,
                measurement_method=measurement_method,
                sex=sex,
            )
        except:
            uk_who_infants_reference = []
        return uk_who_infants_reference
    elif uk_who_reference == UK_WHO_CHILD:
        try:
            uk_who_children_reference = uk_who_lms_array_for_measurement_and_sex(
                age=2.0,
                measurement_method=measurement_method,
                sex=sex,
            )
        except:
            uk_who_children_reference = []
        return uk_who_children_reference
    elif uk_who_reference == UK90_CHILD:
        try:
            uk90_children_reference = uk_who_lms_array_for_measurement_and_sex(
                age=4.0,
                measurement_method=measurement_method,
                sex=sex,
            )
        except:
            uk90_children_reference = []
        return uk90_children_reference
    else:
        raise LookupError(
            f"No data found for {measurement_method} in {sex}s in {uk_who_reference}")
